﻿using System;
using System.Collections.Generic;
using System.IO;
using ICSharpCode.SharpZipLib.Zip.Compression;
using System.Text;
using LZ4;

namespace HpkArchiver
{
    public enum FileType
    {
        Discover,
        Uncompressed,
        ZLIB,
        LZ4,
        Empty
    }
    public class FragmentedFile
    {
        public FileEntry Entry { get; set; }

        public List<Fragment> Fragments { get; set; }
        public Fragment this[int i]
        {
            get { return Fragments[i]; }
            set { Fragments[i] = value; }
        }

        public long Length
        {
            get
            {
                long length = 0;
                for (int i = 0; i < Fragments.Count; i++)
                    length += Fragments[i].Length;
                return length;
            }
        }
        public bool Fragmented
        {
            get
            {
                if (Fragments == null)
                    return false;
                else
                {
                    if (Fragments.Count > 1)
                        return true;
                    else
                        return false;
                }
            }
        }

        public FragmentedFile()
        {
            Fragments = new List<Fragment>();
        }
        public FragmentedFile(List<Fragment> fragments)
        {
            Fragments = fragments;
        }
        public FragmentedFile(ref BinaryReader br, int fragments)
        {
            Fragments = new List<Fragment>();
            for (int i = 0; i < fragments; i++)
            {
                Fragment temp = new Fragment(br.ReadUInt32(), br.ReadInt32());
                if (!temp.IsNull)
                    Fragments.Add(temp);
            }
        }

        /*
        /// <summary>
        /// Extracts a fragmented file from the archive to the destination.  Does NOT decompress.
        /// </summary>
        /// <param name="br">Stream to use to extract the file.</param>
        /// <param name="dest">Path to where the extracted file should be saved.</param>
        /// <returns>Number of errors caught.</returns>
        public int ExtractFile(ref BinaryReader br, string dest)
        {
            int errorscaught = 0;
            try
            {
                byte[] buffer = new byte[4096];
                int read;
                FileStream fs = new FileStream(dest, FileMode.Create, FileAccess.Write);
                for (int i = 0; i < Fragments.Count; i++)
                {
                    br.BaseStream.Seek(Fragments[i].Offset, SeekOrigin.Begin);
                    int length = Fragments[i].Length;
                    while (length > 0)
                    {
                        if (length < buffer.Length)
                            read = br.BaseStream.Read(buffer, 0, length);
                        else
                            read = br.BaseStream.Read(buffer, 0, buffer.Length);
                        length -= read;
                        fs.Write(buffer, 0, read);
                        fs.Flush();
                    }
                }
                fs.Close();
            }
            catch (Exception ex)
            {
                errorscaught++;
                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                sw.WriteLine(dest + "\tCould not extract file: " + ex.Message);
                sw.Flush();
                sw.Close();
            }
            return errorscaught;
        }
        /// <summary>
        /// Extracts a fragmented Zlib compressed file.
        /// </summary>
        /// <param name="br">Stream to use to extract the file.</param>
        /// <param name="dest">Path to where the extracted file should be saved.</param>
        /// <returns>Number of errors caught.</returns>
        public int ExtractZlibFile(ref BinaryReader br, string dest)
        {
            int errorscaught = 0;
            try
            {
                if (Fragments.Count == 0 || Length < 16) // Bug fix for corruption/incomplete/empty file
                {
                    File.Create(dest).Close(); // Make blank file.

                    errorscaught++;
                    StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                    sw.WriteLine(dest + "\tWarning: File too small to decompress, outputting blank file.");
                    sw.Flush();
                    sw.Close();

                    return errorscaught;
                }

                FragmentedBinaryReader fbr = new FragmentedBinaryReader(ref br, Fragments);

                ZlibHeader header = new ZlibHeader();
                byte[] head_bytes = fbr.ReadFragmented(16);
                if (ZlibHeader.MAGIC_CHARS != Encoding.ASCII.GetString(head_bytes, 0, 4))
                {
                    errorscaught++;
                    StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                    sw.WriteLine(dest + "\tExpecting " + ZlibHeader.MAGIC_CHARS + ".");
                    sw.Flush();
                    sw.Close();

                    return errorscaught;
                }
                header.InflatedLength = BitConverter.ToUInt32(head_bytes, 4);
                header.ChunkSize = BitConverter.ToInt32(head_bytes, 8);
                header.OffsetCollection.Add(BitConverter.ToInt32(head_bytes, 12));
                if (header.OffsetCollection[0] != 16)
                {
                    head_bytes = fbr.ReadFragmented(header.OffsetCollection[0] - 16);
                    for (int i = 0; i < head_bytes.Length; i += 4)
                        header.OffsetCollection.Add(BitConverter.ToInt32(head_bytes, i));
                }

                int len = Convert.ToInt32(Length);
                Stack<int> lengths = new Stack<int>(header.OffsetCollection.Count);
                for (int i = header.OffsetCollection.Count - 1; i >= 0; i--)
                {
                    lengths.Push(len - header.OffsetCollection[i]);
                    len -= lengths.Peek();
                }
                len -= header.OffsetCollection[0];


                FileInfo fi = new FileInfo(dest);
                if (!fi.Directory.Exists)
                    fi.Directory.Create();

                byte[] buffer = new byte[4096];
                FileStream compressed_out = new FileStream(dest, FileMode.Create, FileAccess.Write);

                int o = 0;
                int result = 0;
                while (lengths.Count > 0)
                {
                    len = lengths.Pop();

                    byte[] destination = new byte[header.ChunkSize];
                    byte[] source = fbr.ReadFragmented(len);

                    Inflater inflate = new Inflater();

                    result = 0;
                    try
                    {
                        if (source.Length == header.ChunkSize)
                            throw new Exception("Warning: Can't inflate when expected size matches input size."); // Send to catch to be written

                        // attempt to inflate
                        inflate.SetInput(source);
                        result = inflate.Inflate(destination);
                    }
                    catch { result = 0; }; // fall through to the code below
                    if (result == 0)
                    {
                        // dump the chunk to file -- it may not be compressed at all
                        Array.Copy(source, destination, source.Length);
                        result = source.Length;
                    }
                    compressed_out.Write(destination, 0, result);
                    compressed_out.Flush();
                    o++;
                }
                compressed_out.Close();

                if (result == 0)
                {
                    if (header.OffsetCollection.Count == 1 && header.InflatedLength < 32)
                    {
                        StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                        sw.WriteLine(dest + "\tWarning: Could not decompress the file.  Writing raw data.");
                        sw.Flush();
                        sw.Close();

                        FileStream fs = new FileStream(dest, FileMode.Create, FileAccess.Write);
                        br.BaseStream.Seek(Fragments[0].Offset + header.OffsetCollection[0], SeekOrigin.Begin);
                        buffer = br.ReadBytes(Convert.ToInt32(header.InflatedLength));
                        fs.Write(buffer, 0, buffer.Length);
                        fs.Flush();
                        fs.Close();
                    }
                    else
                    {
                        errorscaught++;
                        StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                        sw.WriteLine(dest + "\tCould not decompress the file.  Writing compressed file.");
                        sw.Flush();
                        sw.Close();

                        ExtractFile(ref br, dest);
                    }
                }
            }
            catch (Exception ex)
            {
                errorscaught++;
                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                sw.WriteLine(dest + "\t" + ex.Message);
                sw.Flush();
                sw.Close();
            }

            return errorscaught;
        }
        /// <summary>
        /// Extracts a fragmented LZ4 compressed file.
        /// </summary>
        /// <param name="br">Stream to use to extract the file.</param>
        /// <param name="dest">Path to where the extracted file should be saved.</param>
        /// <returns>Number of errors caught.</returns>
        public int ExtractLZ4File(ref BinaryReader br, string dest)
        {
            int errorscaught = 0;
            try
            {
                if (Fragments.Count == 0 || Length < 16) // Bug fix for corruption/incomplete/empty file
                {
                    File.Create(dest).Close(); // Make blank file.

                    errorscaught++;
                    StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                    sw.WriteLine(dest + "\tWarning: File too small to decompress, outputting blank file.");
                    sw.Flush();
                    sw.Close();

                    return errorscaught;
                }

                FragmentedBinaryReader fbr = new FragmentedBinaryReader(ref br, Fragments);

                LZ4Header header = new LZ4Header();
                byte[] head_bytes = fbr.ReadFragmented(16);
                if (LZ4Header.MAGIC_CHARS != Encoding.ASCII.GetString(head_bytes, 0, 4))
                {
                    errorscaught++;
                    StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                    sw.WriteLine(dest + "\tExpecting " + LZ4Header.MAGIC_CHARS + ".");
                    sw.Flush();
                    sw.Close();

                    return errorscaught;
                }
                header.InflatedLength = BitConverter.ToUInt32(head_bytes, 4);
                header.ChunkSize = BitConverter.ToInt32(head_bytes, 8);
                header.OffsetCollection.Add(BitConverter.ToInt32(head_bytes, 12));
                if (header.OffsetCollection[0] != 16)
                {
                    head_bytes = fbr.ReadFragmented(header.OffsetCollection[0] - 16);
                    for (int i = 0; i < head_bytes.Length; i += 4)
                        header.OffsetCollection.Add(BitConverter.ToInt32(head_bytes, i));
                }

                int len = Convert.ToInt32(Length);
                Stack<int> lengths = new Stack<int>(header.OffsetCollection.Count);
                for (int i = header.OffsetCollection.Count - 1; i >= 0; i--)
                {
                    lengths.Push(len - header.OffsetCollection[i]);
                    len -= lengths.Peek();
                }
                len -= header.OffsetCollection[0];


                FileInfo fi = new FileInfo(dest);
                if (!fi.Directory.Exists)
                    fi.Directory.Create();

                byte[] buffer = new byte[4096];
                FileStream compressed_out = new FileStream(dest, FileMode.Create, FileAccess.Write);

                int o = 0;
                int result = 0;
                while (lengths.Count > 0)
                {
                    len = lengths.Pop();

                    byte[] destination = new byte[header.ChunkSize];
                    byte[] source = fbr.ReadFragmented(len);

                    result = 0;
                    try
                    {
                        if (source.Length == header.ChunkSize)
                            throw new Exception("Warning: Can't inflate when expected size matches input size."); // Send to catch to be written

                        // attempt to inflate
                        result = LZ4Codec.Decode(source, 0, source.Length, destination, 0, destination.Length);
                    }
                    catch { result = 0; }; // fall through to the code below
                    if (result == 0)
                    {
                        // dump the chunk to file -- it may not be compressed at all
                        Array.Copy(source, destination, source.Length);
                        result = source.Length;
                    }
                    compressed_out.Write(destination, 0, result);
                    compressed_out.Flush();
                    o++;
                }
                compressed_out.Close();

                if (result == 0)
                {
                    if (header.OffsetCollection.Count == 1 && header.InflatedLength < 32)
                    {
                        StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                        sw.WriteLine(dest + "\tWarning: Could not decompress the file.  Writing raw data.");
                        sw.Flush();
                        sw.Close();

                        FileStream fs = new FileStream(dest, FileMode.Create, FileAccess.Write);
                        br.BaseStream.Seek(Fragments[0].Offset + header.OffsetCollection[0], SeekOrigin.Begin);
                        buffer = br.ReadBytes(Convert.ToInt32(header.InflatedLength));
                        fs.Write(buffer, 0, buffer.Length);
                        fs.Flush();
                        fs.Close();
                    }
                    else
                    {
                        errorscaught++;
                        StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                        sw.WriteLine(dest + "\tCould not decompress the file.  Writing compressed file.");
                        sw.Flush();
                        sw.Close();

                        ExtractFile(ref br, dest);
                    }
                }
            }
            catch (Exception ex)
            {
                errorscaught++;
                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                sw.WriteLine(dest + "\t" + ex.Message);
                sw.Flush();
                sw.Close();
            }

            return errorscaught;
        }
        */

        /// <summary>Extracts a file from the archive, automatically decompressing it if it is decompressed.</summary>
        /// <param name="br">The stream set to offset of the file.</param>
        /// <param name="dest">Where to extract the file to. Will always be created.</param>
        /// <param name="type">If the type is known, provide it and it will skip discovery.</param>
        /// <returns>Number of errors encountered.</returns>
        public int ExtractFile(ref BinaryReader br, string dest, FileType type = FileType.Discover)
        {
            int errorscaught = 0;
            try
            {
                FileStream fs;
                byte[] buffer = new byte[4096];

                // figure out what we're dealing with if it isn't given
                if (type == FileType.Discover)
                {
                    long length = Length;
                    if (length == 0)
                        type = FileType.Empty;
                    else
                    {
                        if (length < 12)
                            type = FileType.Uncompressed;
                        else
                        {
                            switch(Encoding.ASCII.GetString(br.ReadBytes(4)))
                            {
                                case Static.LZ4:
                                    type = FileType.LZ4;
                                    break;
                                case Static.ZLIB:
                                    type = FileType.ZLIB;
                                    break;
                                default:
                                    type = FileType.Uncompressed;
                                    break;
                            }
                            br.BaseStream.Seek(-4, SeekOrigin.Current);
                        }
                    }
                }

                switch (type)
                {
                    case FileType.Empty:
                        File.Create(dest).Close();
                        break;
                    case FileType.ZLIB:
                    case FileType.LZ4:
                        if (Length < 16) // Bug fix for corruption/incomplete/empty file
                        {
                            File.Create(dest).Close(); // Make blank file.
                            return errorscaught;
                        }

                        FragmentedBinaryReader fbr = new FragmentedBinaryReader(ref br, Fragments);
                        CompressedHeader header = new CompressedHeader();
                        byte[] head_bytes = fbr.ReadFragmented(16);
                        header.MagicChars = Encoding.ASCII.GetString(head_bytes, 0, 4);
                        header.InflatedLength = BitConverter.ToUInt32(head_bytes, 4);
                        header.ChunkSize = BitConverter.ToInt32(head_bytes, 8);
                        header.OffsetCollection.Add(BitConverter.ToInt32(head_bytes, 12));
                        if (header.OffsetCollection[0] != 16)
                        {
                            head_bytes = fbr.ReadFragmented(header.OffsetCollection[0] - 16);
                            for (int i = 0; i < head_bytes.Length; i += 4)
                                header.OffsetCollection.Add(BitConverter.ToInt32(head_bytes, i));
                        }

                        int len = Convert.ToInt32(Length);
                        Stack<int> lengths = new Stack<int>(header.OffsetCollection.Count);
                        for (int i = header.OffsetCollection.Count - 1; i >= 0; i--)
                        {
                            lengths.Push(len - header.OffsetCollection[i]);
                            len -= lengths.Peek();
                        }
                        len -= header.OffsetCollection[0];


                        FileInfo fi = new FileInfo(dest);
                        if (!fi.Directory.Exists)
                            fi.Directory.Create();
                        
                        FileStream compressed_out = new FileStream(dest, FileMode.Create, FileAccess.Write);

                        int o = 0;
                        int result = 0;
                        Inflater inflate = new Inflater();
                        while (lengths.Count > 0)
                        {
                            len = lengths.Pop();

                            byte[] destination = new byte[header.ChunkSize];
                            byte[] source = fbr.ReadFragmented(len);

                            result = 0;
                            try
                            {
                                if (source.Length == header.ChunkSize)
                                    throw new Exception("Warning: Can't inflate when expected size matches input size."); // Send to catch to be written

                                // attempt to inflate
                                if (type == FileType.LZ4)
                                    result = LZ4Codec.Decode(source, 0, source.Length, destination, 0, destination.Length);
                                else
                                {
                                    inflate.SetInput(source);
                                    result = inflate.Inflate(destination);
                                }
                            }
                            catch { result = 0; }; // fall through to the code below
                            if (result == 0)
                            {
                                // dump the chunk to file -- it may not be compressed at all
                                Array.Copy(source, destination, source.Length);
                                result = source.Length;
                            }
                            compressed_out.Write(destination, 0, result);
                            compressed_out.Flush();
                            o++;
                        }
                        compressed_out.Close();

                        if (result == 0)
                        {
                            if (header.OffsetCollection.Count == 1 && header.InflatedLength < 32)
                            {
                                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                                sw.WriteLine(dest + "\tWarning: Could not decompress the file.  Writing raw data.");
                                sw.Flush();
                                sw.Close();

                                fs = new FileStream(dest, FileMode.Create, FileAccess.Write);
                                br.BaseStream.Seek(Fragments[0].Offset + header.OffsetCollection[0], SeekOrigin.Begin);
                                buffer = br.ReadBytes(Convert.ToInt32(header.InflatedLength));
                                fs.Write(buffer, 0, buffer.Length);
                                fs.Flush();
                                fs.Close();
                            }
                            else
                            {
                                errorscaught++;
                                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                                sw.WriteLine(dest + "\tCould not decompress the file.  Writing compressed file.");
                                sw.Flush();
                                sw.Close();

                                ExtractFile(ref br, dest);
                            }
                        }
                        break;
                    default:
                        int read;
                        fs = new FileStream(dest, FileMode.Create, FileAccess.Write);
                        for (int i = 0; i < Fragments.Count; i++)
                        {
                            br.BaseStream.Seek(Fragments[i].Offset, SeekOrigin.Begin);
                            int length = Fragments[i].Length;
                            while (length > 0)
                            {
                                if (length < buffer.Length)
                                    read = br.BaseStream.Read(buffer, 0, length);
                                else
                                    read = br.BaseStream.Read(buffer, 0, buffer.Length);
                                length -= read;
                                fs.Write(buffer, 0, read);
                                fs.Flush();
                            }
                        }
                        fs.Close();
                        break;
                }
            }
            catch (Exception ex)
            {
                errorscaught++;
                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                sw.WriteLine(dest + "\tCould not extract file: " + ex.Message);
                sw.Flush();
                sw.Close();
            }
            return errorscaught;
        }

        public override string ToString()
        {
            if (Entry == null)
            {
                if (Fragments.Count > 1)
                    return Fragments[0].Offset + "," + Length.ToString() + "*";
                else
                    return Fragments[0].ToString();
            }
            else
            {
                if (Fragments.Count > 1)
                    return Entry.Name + " (" + Fragments[0].Offset + "," + Length.ToString() + "*)";
                else
                    return Entry.Name + " (" + Fragments[0].ToString() + ")";
            }
        }
        public override bool Equals(object obj)
        {
            if (obj.GetType() == typeof(FragmentedFile))
            {
                FragmentedFile ff = (FragmentedFile)obj;
                if (ff.Fragments == Fragments)
                    return true;
            }
            return false;
        }
        public override int GetHashCode()
        {
            return Fragments.GetHashCode();
        }
    }
}
