﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using ICSharpCode.SharpZipLib.Zip.Compression;
using LZ4;

namespace HpkArchiver
{
    public delegate void FileSystemEntryProcessedHandler(FragmentType type);
    public delegate void FileSystemCountHandler(int count);
    public delegate void ProcessingFileDatesHandler();
    public delegate void CompletedHandler(TimeSpan duration, string description);

    public class HpkFile
    {
        public event FileSystemEntryProcessedHandler FileSystemEntryProcessed;
        public event FileSystemCountHandler FileSystemCount;
        public event ProcessingFileDatesHandler ProcessingFileDates;
        public event CompletedHandler Completed;

        private int _ErrorsCaught = 0;
        public int ErrorsCaught
        {
            get { return _ErrorsCaught; }
        }

        private int _FilesHandled = 0;
        public int FilesHandled
        {
            get { return _FilesHandled; }
        }

        private int _DirectoriesHandled = 0;
        public int DirectoriesHandled
        {
            get { return _DirectoriesHandled; }
        }

        private Thread _Thread = null;
        public void Abort()
        {
            if (_Thread != null)
                _Thread.Abort();
        }

        private Game _Game = Game.Tropico3;
        public Game Game
        {
            get { return _Game; }
            set { _Game = value; }
        }

        private const int FRAGMENT_LENGTH = 8; // The length, in bytes, of how long each fragment header (offset/length pairs) appears in the file.
        private const int BUFFER_SIZE = 4096;
        private const int COMPRESS_BUFFER_SIZE_ZLIB = 32768;
        private const int COMPRESS_BUFFER_SIZE_LZ4 = 131072;
        private readonly string[] COMPRESS_EXTENSIONS = new string[] { ".lua", ".xml", ".tga", ".dds", ".xtex", ".bin", ".csv", ".hgs" };

        public HpkFile()
        {
        }

        private Header _Header;
        private List<Fragment> _ResidualCollection;
        public long ResidualBytes
        {
            get
            {
                long output = 0;
                for (int i = 0; i < _ResidualCollection.Count; i++)
                    output += _ResidualCollection[i].Length;
                return output;
            }
        }
        private List<FragmentedFile> _FragmentedFileCollection;
        private List<FileEntryExtended> _FileSystemExtendedCollection;

        public void Clear()
        {
            _ErrorsCaught = 0;
            _FilesHandled = 0;
            _DirectoriesHandled = 0;
            _Header = new Header();
            _ResidualCollection = new List<Fragment>();
            _FragmentedFileCollection = new List<FragmentedFile>();
            _FileSystemExtendedCollection = new List<FileEntryExtended>();
        }

        #region " Extract "
        public void StartExtract(string source, string dest)
        {
            _Thread = new Thread(Extract) { Name = "HpkExtract" };
            _Thread.Start(new HpkExtractArgs(source, dest));
        }
        /// <summary>
        /// Extracts contained directory hive out of an HPK archive.
        /// </summary>
        /// <param name="args">Must be an instance of HpkExtractArgs.</param>
        public void Extract(object args)
        {
            Clear();

            HpkExtractArgs arguments = (HpkExtractArgs)args;

            DateTime start = DateTime.Now;

            FileStream fs = new FileStream(arguments.SourceFileName, FileMode.Open, FileAccess.Read);
            BinaryReader br = new BinaryReader(fs);

            #region " inflate if compressed "
            string unpacked = ""; // temp file.
            FileType type = FileType.Discover; // default type
            switch (Encoding.ASCII.GetString(br.ReadBytes(4)))
            {
                case Static.ZLIB:
                    type = FileType.ZLIB;
                    break;
                case Static.LZ4:
                    type = FileType.LZ4;
                    break;
            }
            if (type != FileType.Discover)
            {
                unpacked = Static.TempZippedFile;
                FragmentedFile hpk = new FragmentedFile();
                hpk.Fragments.Add(new Fragment(0, Convert.ToInt32(fs.Length)));
                _ErrorsCaught += hpk.ExtractFile(ref br, unpacked, type);
                br.Close();
                fs.Close();
                fs = new FileStream(unpacked, FileMode.Open, FileAccess.Read);
                br = new BinaryReader(fs);
            }
            #endregion

            #region " Header "
            fs.Seek(0, SeekOrigin.Begin); // reset to beginning of file
            _Header = new Header(br.ReadBytes(Header.LENGTH));
            int filesystementries = _Header.FragmentedFileSystemLength / (FRAGMENT_LENGTH * _Header.FragmentsPerFile);
            #endregion

            #region " Residual "
            fs.Seek(_Header.FragmentsResidualOffset, SeekOrigin.Begin);
            for (int i = 0; i < _Header.FragmentsResidualCount; i++)
                _ResidualCollection.Add(new Fragment(br.ReadBytes(FRAGMENT_LENGTH)));
            #endregion

            FileSystemCount?.Invoke(filesystementries);

            #region " File System "
            fs.Seek(_Header.FragmentedFileSystemOffset, SeekOrigin.Begin);
            for (int i = 0; i < filesystementries; i++)
                _FragmentedFileCollection.Add(new FragmentedFile(ref br, _Header.FragmentsPerFile));

            ReadDirectory(0, arguments.DestinationFolderName, ref br);
            #endregion

            br.Close();
            fs.Close();

            if ((unpacked.Length > 0) && File.Exists(unpacked)) // remove temp file
                Static.TryDeleteFile(unpacked);

            ProcessFileDates(arguments.DestinationFolderName + Static.PlatformSlash() + "_filedates");

            Completed?.Invoke(DateTime.Now - start, "extracted");
        }

        private void ReadDirectory(int entrylink, string workingdir, ref BinaryReader br)
        {
            if (_FragmentedFileCollection[entrylink].Fragments.Count == 0)
            {
                StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                sw.WriteLine("Warning: File at index " + entrylink.ToString() + " skipped due to not having any attached data.");
                sw.Flush();
                sw.Close();
            }
            else
            {
                br.BaseStream.Seek(_FragmentedFileCollection[entrylink][0].Offset, SeekOrigin.Begin);
                byte[] temp = br.ReadBytes(_FragmentedFileCollection[entrylink][0].Length);

                int position = 0;
                while (position < _FragmentedFileCollection[entrylink][0].Length)
                {
                    br.BaseStream.Seek(_FragmentedFileCollection[entrylink][0].Offset + position, SeekOrigin.Begin);

                    FileEntry entry = new FileEntry
                    {
                        FragmentIndex = BitConverter.ToInt32(temp, position)
                    };
                    position += 4;

                    entry.Type = (FragmentType)BitConverter.ToUInt32(temp, position);
                    position += 4;

                    entry.NameLength = BitConverter.ToUInt16(temp, position);
                    position += 2;

                    entry.Name = Encoding.ASCII.GetString(temp, position, entry.NameLength);
                    position += entry.NameLength;

                    string dir = workingdir;
                    if (entry.Type == FragmentType.Directory)
                    {
                        dir += Static.PlatformSlash() + entry.Name;
                        Directory.CreateDirectory(dir);
                        ReadDirectory(entry.FragmentIndex - 1, dir, ref br);

                        _DirectoriesHandled++;
                        FileSystemEntryProcessed?.Invoke(FragmentType.Directory);
                    }
                    else
                    {
                        string dest = dir + Static.PlatformSlash() + entry.Name;
                        if (_FragmentedFileCollection[entry.FragmentIndex - 1].Fragments.Count == 0)
                        {
                            File.Create(dest).Close(); // Make blank file.

                            _ErrorsCaught++;
                            StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                            sw.WriteLine(dest + "\tNo fragments, outputting blank file.");
                            sw.Flush();
                            sw.Close();
                        }
                        else
                        {
                            br.BaseStream.Seek(_FragmentedFileCollection[entry.FragmentIndex - 1][0].Offset, SeekOrigin.Begin);
                            _ErrorsCaught += _FragmentedFileCollection[entry.FragmentIndex - 1].ExtractFile(ref br, dest); // Extract it, method will figure out if it is compressed or not.

                            _FilesHandled++;
                            FileSystemEntryProcessed?.Invoke(FragmentType.File);
                        }
                    }
                    _FragmentedFileCollection[entry.FragmentIndex - 1].Entry = entry;
                }
            }
        }

        private void ProcessFileDates(string filedates)
        {
            FileInfo src = new FileInfo(filedates);
            if (src.Exists)
            {
                ProcessingFileDates?.Invoke();

                string path = src.DirectoryName + Static.PlatformSlash();

                StreamReader sr = new StreamReader(filedates);
                string a, b;
                switch (_Game)
                {
                    case Game.Tropico4:
                    case Game.Tropico5:
                        while (sr.Peek() != -1)
                        {
                            Static.SplitOnce(sr.ReadLine(), '=', out a, out b);
                            try { File.SetLastWriteTimeUtc(path + Static.PlatformPathReverse(a), DateTime.FromFileTime(Convert.ToInt64(b) * 2000)); }
                            catch { } // it is a directory and setting the timestamp on a directory is forbidden
                        }
                        break;
                    default:
                        while (sr.Peek() != -1)
                        {
                            Static.SplitOnce(sr.ReadLine(), '=', out a, out b);
                            try { File.SetLastWriteTimeUtc(path + Static.PlatformPathReverse(a), DateTime.FromFileTime(Convert.ToInt64(b))); }
                            catch { } // it is a directory and setting the timestamp on a directory is forbidden
                        }
                    break;
                }
                sr.Close();

                // This file is now irrelevant.
                src.Delete();
            }
        }
        #endregion

        #region " Archive "
        public void StartArchive(string source, string dest, bool compress)
        {
            _Thread = new Thread(Archive)
            {
                Name = "HpkArchive"
            };
            _Thread.Start(new HpkArchiveArgs(source, dest, compress));
        }
        /// <summary>
        /// Packs contained directory hive into an HPK archive.
        /// </summary>
        /// <param name="args">An instance of HpkArchiveArgs which tells it what to do.</param>
        public void Archive(object args)
        {
            Clear();

            HpkArchiveArgs arguments = (HpkArchiveArgs)args;

            DateTime start = DateTime.Now;

            DiscoverDirectories(new DirectoryInfo(arguments.SourceFolderName));

            FileEntryExtended fse = new FileEntryExtended
            {
                FragmentIndex = _FileSystemExtendedCollection.Count + 1,
                Name = "_filedates",
                Type = FragmentType.File
            };
            fse.NameLength = Convert.ToUInt16(fse.Name.Length);
            _FileSystemExtendedCollection[0].FileIndices.Add(_FileSystemExtendedCollection.Count);
            _FileSystemExtendedCollection.Add(fse);

            FileSystemCount?.Invoke(_FileSystemExtendedCollection.Count - 1);

            FileStream output;
            string packed = "";
            if (arguments.Compress)
            {
                packed = Static.TempZippedFile;
                output = new FileStream(packed, FileMode.Create, FileAccess.ReadWrite);
            }
            else
                output = new FileStream(arguments.DestinationFileName, FileMode.Create, FileAccess.ReadWrite);
            output.Seek(Header.LENGTH, SeekOrigin.Begin); // reserve room for the header which will be written there later
            BinaryWriter bw = new BinaryWriter(output);

            #region " Write Data "
            for (int i = 0; i < _FileSystemExtendedCollection.Count - 1; i++)
            {
                if (_FileSystemExtendedCollection[i].Type == FragmentType.Directory)
                {
                    _FileSystemExtendedCollection[i].Offset = Convert.ToUInt32(output.Position);

                    for (int j = 0; j < _FileSystemExtendedCollection[i].DirIndices.Count; j++)
                        bw.Write(_FileSystemExtendedCollection[_FileSystemExtendedCollection[i].DirIndices[j]].GetBytes());
                    for (int j = 0; j < _FileSystemExtendedCollection[i].FileIndices.Count; j++)
                        bw.Write(_FileSystemExtendedCollection[_FileSystemExtendedCollection[i].FileIndices[j]].GetBytes());
                    bw.Flush();

                    _FileSystemExtendedCollection[i].Length = Convert.ToInt32(output.Position - _FileSystemExtendedCollection[i].Offset);

                    _DirectoriesHandled++;
                    FileSystemEntryProcessed?.Invoke(FragmentType.Directory);
                }
                else
                {
                    if (_FileSystemExtendedCollection[i].Compress)
                        ArchiveCompressedFile(ref bw, i, CompressionAlgorithm);
                    else
                        ArchiveFile(ref bw, i);

                    _FilesHandled++;
                    FileSystemEntryProcessed?.Invoke(FragmentType.File);
                }
            }
            #endregion

            #region " Write _filedates"
            int filedates_index = _FileSystemExtendedCollection[0].FileIndices[_FileSystemExtendedCollection[0].FileIndices.Count - 1];
            _FileSystemExtendedCollection[filedates_index].Offset = Convert.ToUInt32(output.Position);
            StreamWriter sw = new StreamWriter(output);
            switch (_Game)
            {
                case Game.Tropico4:
                case Game.Tropico5:
                    for (int i = 1; i < _FileSystemExtendedCollection.Count - 1; i++)
                    {
                        sw.WriteLine(Static.PlatformPathToUnix(_FileSystemExtendedCollection[i].FullName.Substring(_FileSystemExtendedCollection[0].FullName.Length + 1)) + "=" + Convert.ToInt64(Math.Round((decimal)_FileSystemExtendedCollection[i].FileTime / (decimal)2000, 0)));
                        if (i % 10 == 0)
                            sw.Flush();
                    }
                    break;
                default:
                    for (int i = 1; i < _FileSystemExtendedCollection.Count - 1; i++)
                    {
                        sw.WriteLine(Static.PlatformPathToUnix(_FileSystemExtendedCollection[i].FullName.Substring(_FileSystemExtendedCollection[0].FullName.Length + 1)) + "=" + _FileSystemExtendedCollection[i].FileTime);
                        if (i % 10 == 0)
                            sw.Flush();
                    }
                    break;
            }
            sw.Flush();
            _FileSystemExtendedCollection[filedates_index].Length = Convert.ToInt32(output.Position - _FileSystemExtendedCollection[_FileSystemExtendedCollection.Count - 1].Offset);
            #endregion

            #region " Write Header "
            Header header = new Header
            {
                DataOffset = Header.LENGTH,
                FragmentsPerFile = 1, // Easiest to work with so simplify all down to one.
                Unknown2 = -1,
                FragmentsResidualOffset = 0,
                FragmentsResidualCount = 0,
                Unknown5 = 1,

                FragmentedFileSystemOffset = Convert.ToUInt32(output.Position)
            };
            for (int i = 0; i < _FileSystemExtendedCollection.Count; i++)
            {
                bw.Write(_FileSystemExtendedCollection[i].Offset);
                bw.Write(_FileSystemExtendedCollection[i].Length);
            }
            bw.Flush();
            header.FragmentedFileSystemLength = Convert.ToInt32(output.Position - header.FragmentedFileSystemOffset);

            output.Seek(0, SeekOrigin.Begin);
            bw.Write(header.GetBytes());
            bw.Flush();
            sw.Close();
            bw.Close();
            output.Close();
            #endregion

            Static.TryDeleteFile(Static.TempFile);

            if ((packed.Length > 0) && File.Exists(packed))
            {
                CompressFile(packed, arguments.DestinationFileName);
                Static.TryDeleteFile(packed);
            }

            Completed?.Invoke(DateTime.Now - start, "archived");
        }

        private int DiscoverDirectories(DirectoryInfo directory)
        {
            int index = _FileSystemExtendedCollection.Count;
            FileEntryExtended dir_fse = new FileEntryExtended
            {
                FullName = directory.FullName,
                FragmentIndex = _FileSystemExtendedCollection.Count + 1,
                Name = directory.Name,
                NameLength = Convert.ToUInt16(directory.Name.Length),
                Type = FragmentType.Directory,
                FileTime = directory.LastWriteTimeUtc.ToFileTime()
            };
            _FileSystemExtendedCollection.Add(dir_fse);

            foreach (DirectoryInfo dir in directory.GetDirectories())
                _FileSystemExtendedCollection[index].DirIndices.Add(DiscoverDirectories(dir));

            foreach (FileInfo file in directory.GetFiles())
            {
                FileEntryExtended fse = new FileEntryExtended
                {
                    FullName = file.FullName,
                    FragmentIndex = _FileSystemExtendedCollection.Count + 1,
                    Name = file.Name,
                    NameLength = Convert.ToUInt16(file.Name.Length),
                    Type = FragmentType.File,
                    Length = Convert.ToInt32(file.Length)
                };
                if ((Array.IndexOf<string>(COMPRESS_EXTENSIONS, file.Extension) != -1) && (fse.Length > 255))
                    fse.Compress = true;
                fse.FileTime = file.LastWriteTimeUtc.ToFileTime();
                _FileSystemExtendedCollection[index].FileIndices.Add(_FileSystemExtendedCollection.Count);
                _FileSystemExtendedCollection.Add(fse);
            }

            return index;
        }

        private void ArchiveFile(ref BinaryWriter bw, int index)
        {
            _FileSystemExtendedCollection[index].Offset = Convert.ToUInt32(bw.BaseStream.Position);

            FileStream input = new FileStream(_FileSystemExtendedCollection[index].FullName, FileMode.Open, FileAccess.Read);
            BinaryReader br = new BinaryReader(input);
            int length = _FileSystemExtendedCollection[index].Length;
            byte[] buffer;
            while (length > 0)
            {
                if (length < BUFFER_SIZE)
                {
                    buffer = br.ReadBytes(length);
                    length -= length;
                }
                else
                {
                    buffer = br.ReadBytes(BUFFER_SIZE);
                    length -= BUFFER_SIZE;
                }

                bw.Write(buffer);
                bw.Flush();
            }
            br.Close();
            input.Close();

            _FileSystemExtendedCollection[index].Length = Convert.ToInt32(bw.BaseStream.Position - _FileSystemExtendedCollection[index].Offset);
        }
        private void ArchiveCompressedFile(ref BinaryWriter bw, int index, FileType type = FileType.ZLIB)
        {

            CompressedHeader header = new CompressedHeader();
            if (type == FileType.LZ4)
            {
                header.MagicChars = Static.LZ4;
                header.ChunkSize = COMPRESS_BUFFER_SIZE_LZ4;
            }
            else
            {
                header.MagicChars = Static.ZLIB;
                header.ChunkSize = COMPRESS_BUFFER_SIZE_ZLIB;
            }
            header.ChunkSize = COMPRESS_BUFFER_SIZE_ZLIB;
            header.InflatedLength = Convert.ToUInt32(_FileSystemExtendedCollection[index].Length);

            byte[] buffer;
            byte[] output_buffer = new byte[header.ChunkSize];
            FileStream input = new FileStream(_FileSystemExtendedCollection[index].FullName, FileMode.Open, FileAccess.Read);
            BinaryReader br = new BinaryReader(input);
            int length = _FileSystemExtendedCollection[index].Length;

            FileStream temp = new FileStream(Static.TempFile, FileMode.Create, FileAccess.ReadWrite);

            int result = 0;
            while (length > 0)
            {
                header.OffsetCollection.Add(Convert.ToInt32(temp.Position));

                if (length < header.ChunkSize)
                {
                    buffer = br.ReadBytes(length);
                    length -= length;
                }
                else
                {
                    buffer = br.ReadBytes(header.ChunkSize);
                    length -= header.ChunkSize;
                }

                Deflater deflate = new Deflater(Deflater.DEFAULT_COMPRESSION, false);
                result = 0;
                try
                {
                    if (type == FileType.LZ4)
                        result = LZ4Codec.Encode(buffer, 0, buffer.Length, output_buffer, 0, header.ChunkSize);
                    else
                    {
                        deflate.SetInput(buffer);
                        deflate.Finish();
                        result = deflate.Deflate(output_buffer);
                    }
                }
                catch { };
                if (result == 0)
                {
                    _ErrorsCaught++;
                    StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                    sw.WriteLine(_FileSystemExtendedCollection[index].FullName + "\tCould not compress the file.");
                    sw.Flush();
                    sw.Close();
                    break;
                }
                else if (result == header.ChunkSize) // Fragmented extract won't attempt to decompress this chunk so we have to make sure it is copied as-is.
                    temp.Write(buffer, 0, result);
                else
                    temp.Write(output_buffer, 0, result);
                temp.Flush();
            }
            br.Close();
            input.Close();

            input = null;
            br = null;

            if (result == 0)
                ArchiveFile(ref bw, index);
            else
            {
                _FileSystemExtendedCollection[index].Offset = Convert.ToUInt32(bw.BaseStream.Position);

                int size = header.GetSize();
                for (int i = 0; i < header.OffsetCollection.Count; i++)
                    header.OffsetCollection[i] += size;
                bw.Write(header.GetBytes());

                length = Convert.ToInt32(temp.Length);
                temp.Seek(0, SeekOrigin.Begin);

                br = new BinaryReader(temp);

                while (length > 0)
                {
                    if (length < BUFFER_SIZE)
                    {
                        buffer = br.ReadBytes(length);
                        length -= length;
                    }
                    else
                    {
                        buffer = br.ReadBytes(BUFFER_SIZE);
                        length -= BUFFER_SIZE;
                    }

                    bw.Write(buffer);
                    bw.Flush();
                }
                br.Close();
                temp.Close();

                _FileSystemExtendedCollection[index].Length = Convert.ToInt32(bw.BaseStream.Position - _FileSystemExtendedCollection[index].Offset);
            }
        }
        private void CompressFile(string src, string dest, FileType type = FileType.ZLIB)
        {
            FileInfo fi_src = new FileInfo(src);

            CompressedHeader header = new CompressedHeader();
            if (type == FileType.LZ4)
            {
                header.MagicChars = Static.LZ4;
                header.ChunkSize = COMPRESS_BUFFER_SIZE_LZ4;
            }
            else
            {
                header.MagicChars = Static.ZLIB;
                header.ChunkSize = COMPRESS_BUFFER_SIZE_ZLIB;
            }
            header.InflatedLength = Convert.ToUInt32(fi_src.Length);

            byte[] buffer;
            byte[] output_buffer = new byte[header.ChunkSize];
            FileStream input = new FileStream(fi_src.FullName, FileMode.Open, FileAccess.Read);
            BinaryReader br = new BinaryReader(input);
            int length = Convert.ToInt32(fi_src.Length);

            FileStream temp = new FileStream(Static.TempFile, FileMode.Create, FileAccess.ReadWrite);

            int result = 0;
            Deflater deflate = new Deflater(Deflater.DEFAULT_COMPRESSION, false);
            while (length > 0)
            {
                header.OffsetCollection.Add(Convert.ToInt32(temp.Position));

                if (length < header.ChunkSize)
                {
                    buffer = br.ReadBytes(length);
                    length -= length;
                }
                else
                {
                    buffer = br.ReadBytes(header.ChunkSize);
                    length -= header.ChunkSize;
                }

                result = 0;
                try
                {
                    if (type == FileType.LZ4)
                        result = LZ4Codec.Encode(buffer, 0, buffer.Length, output_buffer, 0, header.ChunkSize);
                    else
                    {
                        deflate.SetInput(buffer);
                        deflate.Finish();
                        result = deflate.Deflate(output_buffer);
                    }
                }
                catch { };
                if (result == 0)
                {
                    _ErrorsCaught++;
                    StreamWriter sw = new StreamWriter(Static.ErrorFile, true);
                    sw.WriteLine(fi_src.FullName + "\tCould not compress the file.");
                    sw.Flush();
                    sw.Close();
                    break;
                }
                else if (result == header.ChunkSize) // Fragmented extract won't attempt to decompress this chunk so we have to make sure it is copied as-is.
                    temp.Write(buffer, 0, result);
                else
                    temp.Write(output_buffer, 0, result);
                temp.Flush();
            }
            br.Close();
            input.Close();

            input = null;
            br = null;

            if (result != 0)
            {
                FileStream output = new FileStream(dest, FileMode.Create, FileAccess.Write);
                BinaryWriter bw = new BinaryWriter(output);

                int size = header.GetSize();
                for (int i = 0; i < header.OffsetCollection.Count; i++)
                    header.OffsetCollection[i] += size;
                bw.Write(header.GetBytes());

                length = Convert.ToInt32(temp.Length);
                temp.Seek(0, SeekOrigin.Begin);

                br = new BinaryReader(temp);

                while (length > 0)
                {
                    if (length < BUFFER_SIZE)
                    {
                        buffer = br.ReadBytes(length);
                        length -= length;
                    }
                    else
                    {
                        buffer = br.ReadBytes(BUFFER_SIZE);
                        length -= BUFFER_SIZE;
                    }

                    bw.Write(buffer);
                    bw.Flush();
                }
                br.Close();
                temp.Close();
                output.Close();
            }

            Static.TryDeleteFile(Static.TempFile);
        }
        #endregion

        public FileType CompressionAlgorithm
        {
            get
            {
                switch (_Game)
                {
                    case Game.VictorVran:
                        return FileType.LZ4;
                    default:
                        return FileType.ZLIB;
                }
            }
        }
    }

    #region " Helper Classes "
    public struct HpkExtractArgs
    {
        public string SourceFileName;
        public string DestinationFolderName;

        public HpkExtractArgs(string source, string dest)
        {
            SourceFileName = source;
            DestinationFolderName = dest;
        }
    }
    public struct HpkArchiveArgs
    {
        public string SourceFolderName;
        public string DestinationFileName;
        public bool Compress;

        public HpkArchiveArgs(string source, string dest, bool compress)
        {
            SourceFolderName = source;
            DestinationFileName = dest;
            Compress = compress;
        }
    }
    #endregion
}
